""" All basic functionality for acia """

from __future__ import annotations

import copy
import logging
import multiprocessing
from functools import partial
from typing import Callable, Iterator

import cv2
import numpy as np
import tqdm
from PIL import Image, ImageDraw
from shapely.geometry import MultiPolygon, Polygon
from tqdm.contrib.concurrent import process_map

from .utils import mask_to_polygons, polygon_to_mask


def unpack(data, function):
    return function(*data)


class Instance:
    """Cell instance based on an image mask and a label"""

    def __init__(self, mask: np.ndarray, frame: int, label: int):
        self.mask = mask
        self.frame = frame
        self.label = label
        self.id = None  # id is unique in an overlay

        self._polygon = None

    @property
    def binary_mask(self):
        return self.mask == self.label

    @property
    def center(self):
        # compute (x,y) center on pixel level

        bin_mask = self.binary_mask

        x = np.median(np.nonzero(np.max(bin_mask, axis=0)))
        y = np.median(np.nonzero(np.max(bin_mask, axis=1)))

        return (x, y)

    @property
    def area(self) -> float:
        """Compute the area inside the contour

        Returns:
            [float]: area
        """
        return np.sum(self.binary_mask)

    def toMask(self, height, width):
        """
        Render contour mask onto new image

        height: height of the image
        width: width of the image
        """
        bin_mask = self.binary_mask
        m_height, m_width = bin_mask.shape
        assert m_height == height
        assert m_width == width

        return bin_mask

    @property
    def polygon(self) -> Polygon:
        if self._polygon is None:
            # TODO: need to get polygon from mask
            self._polygon = mask_to_polygons(self.binary_mask)
            if self._polygon is None:
                print("Error")

        return self._polygon

    def draw(self, image, draw=None, outlineColor=(255, 255, 0), fillColor=None):
        """Draws instance onto an image

        Args:
            image (np.array | PIL.Image): the image to draw onto
            draw (PIL.ImageDraw, optional): Drawing Tool. Defaults to None.
            outlineColor (tuple, optional): Color of the Instance contour. None means no contour is drawn. Defaults to (255, 255, 0).
            fillColor (tuple, optional): Color of the contour fill. Defaults to None (no filling).

        Returns:
            np.array | PIL.Image: The image containing the drawn contour.
        """
        # TODO: make this more efficient
        if draw is None:
            draw = ImageDraw.Draw(image)

        def get_largest(poly):
            if isinstance(poly, MultiPolygon):
                return poly.geoms[np.argmax([p.area for p in poly.geoms])]
            else:
                return poly

        # get the contour coordinates
        coords = np.stack(get_largest(self.polygon).exterior.coords, axis=0).astype(int)
        # draw the polygon
        draw.polygon(tuple(coords.flatten()), outline=outlineColor, fill=fillColor)


class Contour:
    """Class for object contour detection (e.g. Cell object)"""

    def __init__(
        self, coordinates: np.ndarray, score: float, frame: int, id, label=None
    ):
        """Create Contour

        Args:
            coordinates (np.ndarray): coordinates in (x,y) list
            score (float): segmentation score
            frame (int): frame index
            id (any): unique id
            label: class-defining label of the contour
        """
        self.coordinates = np.array(coordinates, dtype=np.float32)
        self.score = score
        self.frame = frame
        self.id = id
        self.label = label

    def _toMask(self, height: int, width: int) -> np.ndarray:
        """
        Render contour mask onto existing image

        img: pillow image
        fillValue: mask values inside the contour
        outlineValues: mask values on the outline (border)
        """
        # perform rasterization into mask
        return polygon_to_mask(self.polygon, height, width)

    def toMask(self, height, width):
        """
        Render contour mask onto new image

        height: height of the image
        width: width of the image
        """
        return self._toMask(height=height, width=width)

    def draw(self, image, draw=None, outlineColor=(255, 255, 0), fillColor=None):
        if draw is None:
            draw = ImageDraw.Draw(image)
        draw.polygon(self.coordinates, outline=outlineColor, fill=fillColor)

    def scale(self, scale: float):
        """Apply scale factor to contour coordinates

        Args:
            scale (float): the multplication factor
        """
        self.coordinates *= scale

    @property
    def center(self):
        return np.array(Polygon(self.coordinates).centroid.coords[0], dtype=np.float32)

    @property
    def area(self) -> float:
        """Compute the area inside the contour

        Returns:
            [float]: area
        """
        return self.polygon.area

    @property
    def polygon(self) -> Polygon:
        return Polygon(self.coordinates)

    def __repr__(self) -> str:
        return self.id


class Overlay:
    """Overlay contains Contours at different frames and provides functionalities iterate and modify them"""

    def __init__(self, contours: list[Contour], frames=None):
        self.contours = contours
        if frames is not None:
            frames = sorted(list(frames))
        self.__frames = frames

        self.cont_lookup = {cont.id: cont for cont in self.contours}

        # set the appropriate counter for the next id
        if len(self.contours) == 0:
            self.id_counter = 0
        else:
            self.id_counter = np.max(list(self.cont_lookup.keys())) + 1

    def add_contour(self, contour: Contour | Instance):
        if isinstance(contour, Instance):
            contour.id = self.id_counter
            self.id_counter += 1

        self.contours.append(contour)
        self.cont_lookup[contour.id] = contour

    def add_contours(self, contours: list[Contour]):
        for cont in contours:
            self.add_contour(cont)

    def __getitem__(self, id):
        return self.cont_lookup[id]

    def __iter__(self):
        return iter(self.contours)

    def __add__(self, other):
        jointContours = self.contours + other.contours
        return Overlay(jointContours)

    def __len__(self):
        return len(self.contours)

    def numFrames(self):
        return len(self.frames())

    def frames(self):
        if self.__frames:
            return self.__frames
        else:
            return np.unique([c.frame for c in self.contours])

    def scale(self, scale: float):
        """Scale the contour with the specified scale factor

           Applies the scale factor to all coordinates individually

        Args:
            scale (float): [description]
        """
        for cont in self.contours:
            cont.scale(scale)

    def croppedContours(self, cropping_parameters: tuple[slice, slice]):
        y, x = cropping_parameters
        miny, maxy, minx, maxx = y.start, y.stop, x.start, x.stop

        crop_rectangle = Polygon(
            [(minx, miny), (maxx, miny), (maxx, maxy), (minx, maxy)]
        )

        def __crop_function_filter(contour: Contour):
            try:
                return crop_rectangle.contains(Polygon(contour.coordinates))
            # TODO: more precise exception catching here!
            # pylint: disable=W0703
            except Exception:
                # if we have problems to convert to shapely polygon, we cannot include it
                logging.warning(
                    "Have to drop Polygon: It cannot be converted into a shapely Polygon."
                )
                return False

        for cont in filter(__crop_function_filter, self.contours):
            new_cont = copy.deepcopy(cont)
            new_cont.coordinates -= np.array([minx, miny])

            yield new_cont

    def timeIterator(self, startFrame=None, endFrame=None, frame_range=None):
        """
        Creates an iterator that returns an Overlay for every frame between starFrame and endFrame

        startFrame: first frame number
        endFrame: last frame number
        """
        if len(self.frames()) == 0:
            yield Overlay([])

        if startFrame is None:
            startFrame = np.min(self.frames())

        if endFrame is None:
            endFrame = np.max(self.frames())

        assert startFrame >= 0
        assert endFrame >= 0
        assert endFrame <= np.max(self.frames())

        it_frames = range(startFrame, endFrame + 1)

        if self.__frames:
            it_frames = sorted(self.__frames)

        # frame for every contour
        frame_information = np.array(
            list(map(lambda cont: cont.frame, self.contours)), dtype=np.int64
        )
        # numpy array of contours (dtype=np.object)
        contour_array = np.array(self.contours)

        # iterate frames
        for frame in it_frames:
            if frame_range and frame not in frame_range:
                continue

            # mask for contour array for this frame
            cont_mask = frame_information == frame
            # filter sub overlay with all contours in the current frame
            yield Overlay(list(contour_array[cont_mask]))

    def toMasks(self, height, width, binary_mask=True) -> list[np.array]:
        """
        Turn the individual overlays into masks. For every time point we create a mask of all contours.

        returns: List of masks (np.array[bool])

        height: height of the image
        width: width of the image
        """
        masks = []
        for timeOverlay in self.timeIterator():
            if binary_mask:
                local_mask = np.zeros((height, width), dtype=bool)
            else:
                # non-binary
                local_mask = np.zeros((height, width), dtype=np.uint16)

            # combine all contours in one mask
            for i, cont in enumerate(timeOverlay):
                mask = cont.toMask(height=height, width=width)
                if not binary_mask:
                    mask = mask.astype(np.uint16) * (
                        i + 1
                    )  # convert into a non-binary mask

                # combine into a single mask
                local_mask = np.maximum(mask, local_mask)

            # append frame mask to list of masks
            masks.append(local_mask)

        return masks

    def draw(
        self,
        image: np.ndarray | Image.Image,
        outlineColor: str | Callable[[Contour], tuple[int]] = None,
        fillColor: str | Callable[[Contour], tuple[int]] = None,
    ):
        """Draw an overly onto an image frame. Hint: overlay should only contain contours for a single frame

        Args:
            image (np.ndarray | Image): Image to draw onto
            outlineColor (str | Callable[[Contour], tuple[int]], optional): Color of the object outlines. If this is a function, the function computes the color for every contour/instance individually. Defaults to None (no contour is drawn).
            fillColor (str | Callable[[Contour], tuple[int]], optional): Fill color of the object. If this is a function, the function computes the color for every contour/instance individually. Defaults to None (no fill). Defaults to None.

        Returns:
            np.ndarray | Image: the updated image object
        """

        if self.numFrames() > 1:
            logging.warning(
                "Drawing overlay onto a frame while the overlay contains instances from multiple frames!"
            )

        is_numpy = isinstance(image, np.ndarray)

        # Deal with numpy or PIL.Image
        if is_numpy:
            # convert into rgb PIL image
            if len(image.shape) == 2:
                image = cv2.cvtColor(image, cv2.COLOR_GRAY2RGB)
            image = Image.fromarray(image)

        imdraw = ImageDraw.Draw(image)
        for timeOverlay in self.timeIterator():
            for cont in timeOverlay:
                oc_local = outlineColor
                fc_local = fillColor

                # compute the contour color for the object
                if oc_local and isinstance(oc_local, Callable):
                    oc_local = oc_local(cont)
                # compute the fill color for the object
                if fc_local and isinstance(fc_local, Callable):
                    fc_local = fc_local(cont)

                cont.draw(image, outlineColor=oc_local, fillColor=fc_local, draw=imdraw)

        if is_numpy:
            # return the numpy version
            return np.asarray(image)
        else:
            # return the PIL image
            return image


class BaseImage:
    """Base class for an image from an image source"""

    @property
    def raw(self):
        raise NotImplementedError("Please implement this function!")

    @property
    def num_channels(self):
        raise NotImplementedError()

    def get_channel(self, channel: int):
        raise NotImplementedError()


class Processor:
    """Base class for a processor"""


class ImageSequenceSource:
    """Base class for an image sequence source (e.g. Tiff, OMERO, png, ...)"""

    @property
    def num_channels(self) -> int:
        raise NotImplementedError()

    def get_frame(self, frame: int) -> BaseImage:
        raise NotImplementedError()


class RoISource:
    """Base class for a RoI source (e.g. tiff metadata, OMERO, json, ...)"""


class ImageRoISource:
    """
    Contains both, the image and the RoI Source. Provides a joint iterator
    """

    def __init__(self, imageSource: ImageSequenceSource, roiSource: RoISource):
        self.imageSource = imageSource
        self.roiSource = roiSource

    def __iter__(self) -> Iterator[tuple[np.array, Overlay]]:
        return zip(iter(self.imageSource), iter(self.roiSource))

    def __len__(self):
        return min(len(self.imageSource), len(self.roiSource))

    def apply_parallel(self, function, num_workers=None):
        if num_workers is None:
            num_workers = int(np.floor(multiprocessing.cpu_count() * 2 / 3))

        return process_map(function, self, max_workers=num_workers, chunksize=4)

    def apply_parallel_star(self, function, num_workers=None):
        if num_workers is None:
            num_workers = int(np.floor(multiprocessing.cpu_count() * 2 / 3))

        return process_map(
            partial(unpack, function=function),
            self,
            max_workers=num_workers,
            chunksize=4,
        )

    def apply(self, function):
        def limit():
            for _, el in enumerate(self):
                yield el

        return list(tqdm.tqdm(map(function, limit())))

    def apply_star(self, function):
        return list(tqdm.tqdm(map(partial(unpack, function=function), self)))



""" Filters for segmentating overlay objects"""

from functools import partial
from typing import Tuple

import cv2
import numpy as np
import shapely
import shapely.affinity
import tqdm.auto as tqdm
from rtree import index
from shapely.geometry import Polygon
from shapely.validation import make_valid

from acia.base import Overlay


def bbox_to_rectangle(bbox: Tuple[float]):
    minx, miny, maxx, maxy = bbox
    return Polygon([(minx, miny), (maxx, miny), (maxx, maxy), (minx, maxy)])


class NMSFilter:
    """Non-maximum supression filter based on contours"""

    @staticmethod
    def filter(overlay: Overlay, iou_thr=0.1, mode="iou") -> Overlay:
        prefiltered_contours = [
            cont for cont in overlay.contours if len(cont.coordinates) >= 3
        ]

        # sort contours by their score (lowest first)
        sorted_contours = sorted(prefiltered_contours, key=lambda c: c.score)
        # make (valid) shapely polygons
        polygons = [
            make_valid(shapely.geometry.polygon.Polygon(contour.coordinates))
            for contour in sorted_contours
        ]

        keep_list = []

        # build an rtree with bounding boxes
        idx = index.Index()
        for i, p_i in enumerate(polygons):
            minx, miny, maxx, maxy = p_i.bounds
            left = minx
            right = maxx
            top = maxy
            bottom = miny
            idx.insert(i, (left, bottom, right, top))

        for i, p_i in tqdm.tqdm(enumerate(polygons), total=len(polygons)):
            keep = True

            # zero area stuff is not considered
            if p_i.area <= 0:
                keep = False
                keep_list.append(keep)
                continue

            # get the intersection candidates by querying the rtree (overlapping bboxes)
            minx, miny, maxx, maxy = p_i.bounds
            left = minx
            right = maxx
            top = maxy
            bottom = miny
            candidate_idx_list = idx.intersection((left, bottom, right, top))

            candidate_idx_list = list(
                filter(
                    partial(
                        lambda index, loop_index, sorted_contours: index > loop_index
                        and sorted_contours[loop_index].frame
                        == sorted_contours[index].frame,
                        loop_index=i,
                        sorted_contours=sorted_contours,
                    ),
                    candidate_idx_list,
                )
            )

            # for those candidates we will compute the intersections in details
            for j in candidate_idx_list:
                p_j = polygons[j]

                # compute iou
                if mode == "i":
                    iou = p_i.intersection(p_j).area / p_i.area  # (p_i.union(p_j).area)
                elif mode == "iou":
                    iou = p_i.intersection(p_j).area / (p_i.union(p_j).area)
                # compare to threshold
                if iou >= iou_thr:
                    # if exceeding iou drop this cell detection
                    # print("iou: %.2f" % iou)
                    keep = False
                    break

            keep_list.append(keep)

        overlay = Overlay(
            [cont for i, cont in enumerate(sorted_contours) if keep_list[i]]
        )

        return overlay


class SizeFilter:
    """Filter by contour area"""

    @staticmethod
    def filter(overlay: Overlay, min_area, max_area) -> Overlay:
        """Filter an overlay based on contour sizes

        Args:
            overlay (Overlay): the overlay to filter
            min_area ([type]): minimum area of a contour
            max_area ([type]): maximum area of a contour

        Returns:
            Overlay: the filtered overlay
        """
        contour_shapes = [Polygon(cont.coordinates) for cont in overlay.contours]
        result_overlay = Overlay([])
        for cont, shape in zip(overlay.contours, contour_shapes):
            area = shape.area

            if min_area < area < max_area:
                result_overlay.add_contour(cont)

        return result_overlay


class EllipsoidFilter:
    """Contour filter for ellipsoid shape"""

    @staticmethod
    def filter(
        overlay: Overlay, min_width_height_ratio, max_width_height_ratio
    ) -> Overlay:
        result_overlay = Overlay([])
        for cont in overlay.contours:
            if len(cont.coordinates) < 5:
                # need 5 points for ellipsoid fit
                continue
            ellipse = cv2.fitEllipse(cont.coordinates)

            center, (width, height), rotation = ellipse

            width_height_ratio = width / height

            # from shapely.figures import SIZE, GREEN, GRAY, set_limits

            # Let create a circle of radius 1 around center point:
            circ = shapely.geometry.Point(center).buffer(1)

            # Let create the ellipse along x and y:
            ell = shapely.affinity.scale(circ, width / 2, height / 2)

            # Let rotate the ellipse (clockwise, x axis pointing right):
            ellr = shapely.affinity.rotate(ell, rotation)

            # create shapely shape from contour coordinates
            shape = Polygon(cont.coordinates)
            # create minimal rotated rectangle
            min_rect = shape.minimum_rotated_rectangle

            rect_area_error = np.abs(shape.area - min_rect.area) / shape.area
            ellipse_area_error = np.abs(ellr.area - shape.area) / shape.area

            if min_width_height_ratio <= width_height_ratio <= max_width_height_ratio:
                if ellipse_area_error < rect_area_error:
                    # if an ellipse can better explain the cell detection than a rectangle
                    result_overlay.add_contour(cont)

        return result_overlay



"""Module for testing visualization module"""

import os
import unittest
from pathlib import Path

import numpy as np

from acia.segm.local import InMemorySequenceSource
from acia.segm.output import renderVideo
from acia.viz import VideoExporter, VideoExporter2


class TestVideoGen(unittest.TestCase):
    """Testing scalebar and time overlay"""

    def _test_encoding(self, video_file: Path, codec):
        """Random images to video"""

        images = np.random.randint(0, 255, size=(10, 200, 200, 3), dtype=np.uint8)

        with VideoExporter(str(video_file.absolute()), 3, codec=codec) as ve:
            for image in images:
                ve.write(image)

        # Check that video file exists
        self.assertTrue(video_file.exists)
        # Check that it has some content (>10 KB)
        self.assertTrue(os.path.getsize(video_file) > 10 * 1024)

    def _test_encoder(self, encoder):
        images = np.random.randint(0, 255, size=(20, 200, 200, 3), dtype=np.uint8)

        with encoder as ve:
            for image in images:
                ve.write(image)

        # Check that video file exists
        self.assertTrue(encoder.filename.exists)
        # Check that it has some content (>10 KB)
        self.assertTrue(os.path.getsize(encoder.filename) > 10 * 1024)

    def test_vp9(self):
        """Encode video in vp9"""
        self._test_encoding(Path("test.mp4"), codec="vp09")

    def test_avi(self):
        """Encode video in avi"""
        self._test_encoding(Path("test.avi"), codec="MJPG")

    def test_avi_fail(self):
        """Encode with wrong codec parameter"""
        with self.assertRaises(TypeError):
            self._test_encoding(Path("test.avi"), codec="JPG")

    def test_new_vp9(self):
        self._test_encoder(VideoExporter2.default_vp9(Path("testvp9.mp4"), 3))

    def test_new_h264(self):
        self._test_encoder(VideoExporter2.default_h264(Path("testh264.mp4"), 3))

    def test_new_h265(self):
        self._test_encoder(VideoExporter2.default_h265(Path("testh265.mp4"), 3))


class TestVideoRendering(unittest.TestCase):
    """Testing video rendering with different codecs"""

    def _test_rendering(self, filepath: Path, codec: str, framerate: int):
        """Test video rendering with specific parameters

        Args:
            filepath (Path): video file path
            codec (str): codec name
            framerate (int): video framerate (fps)
        """

        # random images
        images = np.random.randint(0, 255, size=(20, 200, 200, 3), dtype=np.uint8)
        source = InMemorySequenceSource(images)

        # remove existing files
        filepath = Path(filepath)
        if filepath.exists():
            os.remove(filepath)

        # render videos
        renderVideo(source, filename=filepath, framerate=framerate, codec=codec)

        # Check that video file exists
        self.assertTrue(filepath.exists)
        # Check that it has some content (>10 KB)
        self.assertTrue(os.path.getsize(filepath) > 10 * 1024)

    def test_vp9(self):
        """Encode video in vp9"""
        self._test_rendering("test.mp4", codec="vp09", framerate=10)

    def test_avi(self):
        """Encode video in mjpg"""
        self._test_rendering("test.avi", codec="MJPG", framerate=10)

    def test_avi_fail(self):
        """Encode with wrong codec parameter"""
        with self.assertRaises(ValueError):
            self._test_rendering("test.avi", codec="JPG", framerate=10)

    def test_h264(self):
        """Encode video in h264"""
        self._test_rendering("testh264.mp4", codec="h264", framerate=10)

    def test_new_h265(self):
        """Encode video in h265"""
        self._test_rendering("testh265.mp4", codec="h265", framerate=10)


if __name__ == "__main__":
    unittest.main()



""" Utils for segmentation testing """

import unittest

import numpy as np
from shapely.geometry import Polygon

from acia.segm.utils import compute_indices
from acia.utils import mask_to_polygons, polygon_to_mask


class TestIndexing(unittest.TestCase):
    """Test the linearization of z and t stacks"""

    def test_both(self):

        setup = dict(size_t=4, size_z=4)

        self.assertEqual(compute_indices(0, **setup), (0, 0))
        self.assertEqual(compute_indices(1, **setup), (0, 1))
        self.assertEqual(compute_indices(8, **setup), (2, 0))
        self.assertEqual(compute_indices(3, **setup), (0, 3))
        self.assertEqual(compute_indices(10, **setup), (2, 2))

    def test_only_t(self):
        setup = dict(size_t=50, size_z=1)

        for i in range(setup["size_t"]):
            self.assertEqual(compute_indices(i, **setup), (i, 0))

    def test_only_z(self):
        setup = dict(size_t=1, size_z=50)

        for i in range(setup["size_z"]):
            self.assertEqual(compute_indices(i, **setup), (0, i))


class TestMaskPolygon(unittest.TestCase):
    """Test consistent mask <-> polygon conversions"""

    def test_realistic_polygon_mask(self):
        """Test: polygon -> mask -> polygon -> mask transformation. Check wheter the area stays persistent"""

        polygon = Polygon(
            [
                [87.0, 312.0],
                [86.0, 313.0],
                [85.0, 313.0],
                [83.0, 315.0],
                [83.0, 317.0],
                [82.0, 318.0],
                [82.0, 320.0],
                [81.0, 321.0],
                [81.0, 324.0],
                [80.0, 325.0],
                [80.0, 329.0],
                [79.0, 330.0],
                [79.0, 335.0],
                [78.0, 336.0],
                [78.0, 341.0],
                [77.0, 342.0],
                [77.0, 349.0],
                [78.0, 350.0],
                [78.0, 351.0],
                [81.0, 354.0],
                [86.0, 354.0],
                [87.0, 353.0],
                [88.0, 353.0],
                [89.0, 352.0],
                [89.0, 351.0],
                [90.0, 350.0],
                [90.0, 349.0],
                [91.0, 348.0],
                [91.0, 344.0],
                [92.0, 343.0],
                [92.0, 339.0],
                [93.0, 338.0],
                [93.0, 333.0],
                [94.0, 332.0],
                [94.0, 328.0],
                [95.0, 327.0],
                [95.0, 324.0],
                [96.0, 323.0],
                [96.0, 316.0],
                [93.0, 313.0],
                [92.0, 313.0],
                [91.0, 312.0],
            ]
        )

        mask = polygon_to_mask(polygon, 500, 500)

        mask_area = np.sum(mask)
        polgon_area = polygon.area

        # np.testing.assert_array_equal(polygon.bounds, mask_bounds(mask))

        self.assertEqual(mask_area, polgon_area)

        re_poly = mask_to_polygons(mask)

        self.assertEqual(polygon.area, re_poly.area)
        np.testing.assert_array_equal(polygon.bounds, re_poly.bounds)

        re_mask = polygon_to_mask(re_poly, 500, 500)

        self.assertEqual(np.sum(mask), np.sum(re_mask))
        np.testing.assert_array_equal(mask, re_mask)

    @staticmethod
    def tets_mask_poly_iter():
        """Testing consistency of multiple mask -> polygon -> mask ... transformations"""

        masks = [np.load("tests/resources/mask.npy")]

        height, width = masks[0].shape
        polygons = [mask_to_polygons(masks[-1])]

        num_iters = 5
        for _ in range(num_iters):
            # convert polygon -> mask and mask -> polygon
            masks.append(polygon_to_mask(polygons[-1], height, width))
            polygons.append(mask_to_polygons(masks[-1]))

            # consisteny with first entry
            np.testing.assert_array_equal(masks[0], masks[-1])

            np.testing.assert_array_equal(polygons[0].centroid, polygons[-1].centroid)
            np.testing.assert_almost_equal(polygons[0].area, polygons[-1])

    def test_simple_mask_to_polygon(self):
        """Test conversion of a simple polygon to a mask and back"""

        polygon = Polygon([[0, 0], [10, 0], [10, 10], [0, 10]])

        polygon_area = polygon.area

        mask = polygon_to_mask(polygon, 100, 100)

        mask_area = np.sum(mask)

        self.assertEqual(polygon_area, mask_area)

        re_polygon = mask_to_polygons(mask)

        self.assertEqual(polygon.area, re_polygon.area)
        self.assertEqual(polygon.centroid, re_polygon.centroid)

        np.testing.assert_array_almost_equal(re_polygon.centroid, [5, 5])


if __name__ == "__main__":
    unittest.main()



"""Utils for segmentation data handling"""

from typing import Tuple

import numpy as np
from shapely.geometry import Polygon

from acia.base import Contour
from acia.utils import pairwise_distances


def compute_indices(frame: int, size_t: int, size_z: int) -> Tuple[int, int]:
    """Compute t and z values from a linearized frame number

    Args:
        frame (int): the linearized frame index
        size_t (int): the total size of the t dimension
        size_z (int): the total size of the z dimension

    Returns:
        (Tuple[int, int]): tuple of (t,z) indices
    """

    if size_t > 1 and size_z > 1:
        t = int(np.floor(frame / size_t))
        z = frame % size_t
    elif size_t > 1:
        t = frame
        z = 0
    elif size_z >= 1:
        t = 0
        z = frame
    elif size_t == 1 and size_z == 1:
        t = 0
        z = 0
    else:
        raise ValueError("This state should not be reachable!")

    return t, z


def length_and_area(contour: Contour) -> Tuple[float, float]:
    """Compute length and area of a contour object (in pixel coordinates)

    Args:
        contour (Contour): contour object

    Returns:
        tuple[float, float]: length and area of the contour
    """

    polygon = Polygon(contour.coordinates)

    length = np.max(
        pairwise_distances(np.array(polygon.minimum_rotated_rectangle.exterior.coords))
    )
    return length, polygon.area



""" Test acia base functionality"""

import time
import unittest

import numpy as np

from acia.base import Contour, Instance
from acia.segm.formats import load_ctc_segmentation_native
from acia.tracking.formats import ctc_track_graph, load_ctc_tracklet_graph


class TestContour(unittest.TestCase):
    """Test contour functionality"""

    def test_center(self):
        contour = [[0, 0], [1, 0], [1, 1], [0, 1]]

        self.assertTrue(contour is not None)

        # simple contour
        np.testing.assert_array_equal(
            Contour(contour, 0, 0, 0).center, np.array([0.5, 0.5], dtype=np.float32)
        )

        # unequal point sampling
        contour = [[0, 0], [0.5, 0], [1, 0], [1, 1], [0, 1]]
        np.testing.assert_array_equal(
            Contour(contour, 0, 0, 0).center, np.array([0.5, 0.5], dtype=np.float32)
        )

    def test_rasterization(self):
        """Make sure that contour to mask rasterization preserves area"""

        contours = [[[0, 0], [10, 0], [10, 10], [0, 10]]]

        for coordinates in contours:
            cont = Contour(coordinates, -1, 0, -1)
            mask = cont.toMask(40, 40)

            self.assertEqual(cont.area, np.sum(mask))


class TestInstance(unittest.TestCase):
    """Test contour functionality"""

    def test_center(self):
        mask = np.array(
            [
                [0, 0, 0, 0, 0],
                [0, 0, 3, 3, 0],
                [0, 3, 3, 3, 3],
                [0, 2, 2, 0, 0],
            ],
            dtype=np.uint8,
        )

        instance = Instance(mask, 0, 3)

        self.assertEqual(instance.area, 6)

        poly = instance.polygon

        self.assertEqual(poly.area, 6)

        instance = Instance(mask, 0, 2)
        self.assertEqual(instance.area, 2)

        poly = instance.polygon
        self.assertEqual(poly.area, 2)

    def test_load_native(self):
        print("native")
        t_start = time.time()
        ov = load_ctc_segmentation_native("00_GT/SEG")
        t_end = time.time()
        print(len(ov))
        print("Loading time (s) ", t_end - t_start)

        t_start = time.time()
        _ = [cont.area for cont in ov]
        t_end = time.time()

        ids = [cont.id for cont in ov]

        self.assertEqual(len(ids), len(set(ids)))

        print("Area computation time (s)", t_end - t_start)

    def test_load_tracking_native(self):

        print("native")
        t_start = time.time()
        ov = load_ctc_segmentation_native("00_GT/TRA")
        t_end = time.time()
        print(len(ov))
        print("Load overlay time (s) ", t_end - t_start)

        t_start = time.time()
        tracklet_graph = load_ctc_tracklet_graph("00_GT/TRA/man_track.txt")
        t_end = time.time()
        print("Load tracking time (s) ", t_end - t_start)

        labels = np.unique([cont.label for cont in ov])
        nodes = set(tracklet_graph.nodes)

        # make sure that the labels from the tracking annotations are equal to the tracklet nodes
        self.assertEqual(set(labels), nodes)

        # make the tracking graph
        track_graph = ctc_track_graph(ov, tracklet_graph)

        self.assertEqual(len(ov), track_graph.number_of_nodes())

        for node_id in track_graph.nodes:
            for succ_id in track_graph.successors(node_id):
                self.assertTrue(ov[node_id].frame <= ov[succ_id].frame - 1)

            for pred_id in track_graph.predecessors(node_id):
                self.assertTrue(ov[node_id].frame >= ov[pred_id].frame + 1)


if __name__ == "__main__":
    unittest.main()



"""Unit test cases for tracking subsampling"""

import unittest
from typing import List

import networkx as nx
import numpy as np

from acia.base import Contour, Overlay
from acia.tracking import TrackingSource, TrackingSourceInMemory, subsample_tracking


class TestSubsampling(unittest.TestCase):
    """Test the linearization of z and t stacks"""

    @staticmethod
    def linear_track(
        length, frame_start, id_start=None, successors: List[TrackingSource] = None
    ):

        if successors is None:
            successors = []

        if id_start is None:
            id_start = 0
            for succ in successors:
                id_start = np.max([id_start, *[cont.id for cont in succ.overlay]]) + 1

        ids = [id_start + i for i in range(length)]
        frames = [frame_start + i for i in range(length)]

        overlay = Overlay(
            [Contour(None, -1, frame, id) for id, frame in zip(ids, frames)]
        )
        tracking_graph = nx.DiGraph()
        tracking_graph.add_nodes_from([cont.id for cont in overlay])
        tracking_graph.add_edges_from(
            [(a.id, b.id) for a, b in zip(overlay.contours, overlay.contours[1:])]
        )

        linear_ts = TrackingSourceInMemory(overlay, tracking_graph)

        for succ in successors:
            # merge the two trackings
            linear_ts.merge(succ)

            # add edge between last and first overlay contour
            linear_ts.tracking_graph.add_edge(
                overlay.contours[-1].id, succ.overlay.contours[0].id
            )

        return linear_ts

    def test_linear(self):
        """Test subsampling of a non-dividing contour"""

        # create tracking source
        tsim = TestSubsampling.linear_track(10, 0)

        # subsample tracking
        tr_source = subsample_tracking(tsim, 2)

        self.assertEqual(tr_source.tracking_graph.number_of_nodes(), 5)
        self.assertEqual(tr_source.tracking_graph.number_of_edges(), 4)

        self.assertSetEqual(
            set(map(lambda c: c.id, tr_source.overlay)), {0, 2, 4, 6, 8}
        )

    def test_division(self):
        """Test subsampling of a dividing contour"""

        # create tracking source
        tsim = TestSubsampling.linear_track(
            5,
            0,
            successors=[
                TestSubsampling.linear_track(3, 5, id_start=0),
                TestSubsampling.linear_track(5, 5, id_start=3),
            ],
        )

        # subsample tracking
        tr_source = subsample_tracking(tsim, 3)

        self.assertEqual(tr_source.tracking_graph.number_of_nodes(), 5)
        self.assertEqual(tr_source.tracking_graph.number_of_edges(), 4)
        self.assertSetEqual(
            set(map(lambda c: c.id, tr_source.overlay)), {8, 11, 1, 4, 7}
        )
        self.assertEqual(tr_source.tracking_graph.out_degree(11), 2)



"""Utilities for tracking
"""

import networkx as nx

from acia.tracking.output import CTCTrackingHelper


def life_cycle_lineage(tr_graph: nx.DiGraph) -> nx.DiGraph:
    """Compresses populated lineage to life cycle lineage (one node per cell cycle)

    Args:
        tr_graph (nx.DiGraph): populated tracking graph

    Returns:
        nx.DiGraph: Life cycle lineage with cell cylces as nodes
    """

    # compute the life-cycles of individual cells
    life_cycles = CTCTrackingHelper.compute_life_cycles(tr_graph)
    # create lookup (cont id --> life cycle index)
    life_cycle_lookup = CTCTrackingHelper.create_life_cycle_lookup(life_cycles)
    # contour_lookup = {cont.id: cont for cont in overlay}

    lc_graph = nx.DiGraph()

    # add all the nodes
    lc_graph.add_nodes_from(range(len(life_cycles)))

    # set the "cycle" property to contain the populated life cycle nodes
    for i, life_cycle in enumerate(life_cycles):
        lc_graph.nodes[i]["cycle"] = life_cycle

    # iterate over life_cycles
    for lc_id, lc in enumerate(life_cycles):
        start = lc[0]

        # extract parents from populated tracking
        parents = tr_graph.predecessors(start)

        for parent in parents:
            # get the parent life_cycle
            parent_lc_id = life_cycle_lookup[parent]

            # establish an edge between parent and child
            lc_graph.add_edge(parent_lc_id, lc_id)

    # set "start_frame" and "end_frame" for every node in the life cycle graph
    for node in lc_graph:
        lc = lc_graph.nodes[node]["cycle"]

        lc_graph.nodes[node]["start_frame"] = tr_graph.nodes[lc[0]]["frame"]
        lc_graph.nodes[node]["end_frame"] = tr_graph.nodes[lc[-1]]["frame"]

    return lc_graph



"""Processors to apply remote models for segmentation"""

import json
import logging
import os
from io import BytesIO
from itertools import chain, islice
from typing import List
from urllib.parse import urlparse

import numpy as np
import requests
import tqdm.auto as tqdm
from PIL import Image
from retry import retry

from acia.base import Contour, ImageSequenceSource, Overlay, Processor


class OnlineModel(Processor):
    """
    The model is not running locally on the computer but in a remote location
    """

    def __init__(self, url: str, username=None, password=None, timeout=30):
        """
        url: remote model executer (can also contain a port definition)
        username: username
        password: password
        """
        self.url = url
        self.username = username
        self.password = password
        self.timeout = timeout

        # try to parse port from url
        self.port = urlparse(url).port

        if not self.port:
            # autodetermine port from protocol
            scheme = urlparse(url).scheme
            if scheme == "http":
                self.port = 80
            elif scheme == "https":
                self.port = 443
            else:
                logging.warning(
                    'Could not determine port! Did you specify "http://" or "https://" at the beginning of your url?'
                )

    def predict(self, source: ImageSequenceSource, params=None):
        if params is None:
            params = {}

        contours = []

        # iterate over images from image source
        for frame, image in enumerate(tqdm.tqdm(source)):

            # convert image into a binary png stream
            byte_io = BytesIO()
            Image.fromarray(image).save(byte_io, "png")
            byte_io.seek(0)

            # pack this into form data
            multipart_form_data = {
                "data": ("data.png", byte_io, "image/png"),
            }

            # send a request to the server
            response = requests.post(
                self.url, files=multipart_form_data, params=params, timeout=self.timeout
            )

            # raise an error if the response is not as expected
            if response.status_code != 200:
                raise ValueError(
                    f"HTTP request for prediction not successful. Status code: {response.status_code}"
                )

            body = response.json()

            content = body

            for detection in content:
                # label = detection['label']
                contour_lists = detection["contours"][0]
                contour = list(zip(contour_lists["x"], contour_lists["y"]))
                score = detection["score"]

                contours.append(Contour(contour, score, frame, -1))

        return Overlay(contours)

    @staticmethod
    def parseContours(response_body) -> List[Contour]:
        pass


class ModelDescriptor:
    """Describing all parameters to run a segmentation model from a git repository"""

    def __init__(self, repo: str, entry_point: str, version: str, parameters=None):
        if parameters is None:
            parameters = {}

        self.repo = repo
        self.entry_point = (entry_point,)
        self.version = version
        self.parameters = parameters


def batch(iterable, size):
    """Make iterable over packages of a certain size

    Args:
        iterable (_type_): source iterable
        size (_type_): size of the batches

    Yields:
        _type_: iterable over individual batches
    """

    sourceiter = iter(iterable)
    while True:
        batchiter = islice(sourceiter, size)
        try:
            logging.debug("Next batch...")
            # get first element to known when iterator is at end (will throw StopIteration)
            first_element = next(batchiter)
            if size == 1:
                # single sized batches shall also be lists
                yield [first_element]
            else:
                # chain longer batches
                yield chain([first_element], batchiter)
        except StopIteration:
            # we have reached the end of the iterator
            # --> leave loop
            return


class FlexibleOnlineModel(Processor):
    """
    The model is not running locally on the computer but in a remote location
    """

    def __init__(
        self, executorUrl: str, modelDesc: ModelDescriptor, timeout=600, batch_size=1
    ):
        self.url = executorUrl
        self.timeout = timeout
        self.modelDesc = modelDesc
        self.batch_size = batch_size

        # try to parse port from url
        self.port = urlparse(self.url).port

        if not self.port:
            # autodetermine port from protocol
            scheme = urlparse(self.url).scheme
            if scheme == "http":
                self.port = 80
            elif scheme == "https":
                self.port = 443
            else:
                logging.warning(
                    'Could not determine port! Did you specify "http://" or "https://" at the beginning of your url?'
                )

        # get username from environment for statistics
        # 1. try to get jupyter user, then try to readout other usernames
        self.username = (
            os.environ.get("JUPYTERHUB_USER", None)
            or os.environ.get("USER", None)
            or os.environ.get("USERNAME", None)
        )

    def predict(self, source: ImageSequenceSource, params=None):
        if params is None:
            params = {}

        contours = []

        # Segmentation method parameters
        additional_parameters = dict(self.modelDesc.parameters)
        additional_parameters.update(**params)

        # http request parameters
        params = dict(
            repo=self.modelDesc.repo,
            entry_point=self.modelDesc.entry_point,
            version=self.modelDesc.version,
            username=self.username,
            parameters=json.dumps(additional_parameters),
        )

        # collect all the frame indices that we use for prediction
        all_frames = []

        # Do batch prediction
        for local_batch in batch(enumerate(tqdm.tqdm(source)), self.batch_size):
            local_batch = np.array(list(local_batch), dtype=object)

            frames = local_batch[:, 0]
            all_frames += list(frames)
            images = map(lambda img: img.raw, local_batch[:, 1])

            contours += self.predict_batch(frames, images, params)

        for i, cont in enumerate(contours):
            cont.id = i

        # create new overlay based on all contours and frame numbers
        return Overlay(contours, frames=all_frames)

    def predict_single(self, frame_id, image, params):
        contours = []

        # convert image into a binary png stream
        byte_io = BytesIO()
        Image.fromarray(image).save(byte_io, "png")
        byte_io.seek(0)

        # pack this into form data
        multipart_form_data = {
            "file": ("data.png", byte_io, "image/png"),
        }

        # send a request to the server
        response = requests.post(
            self.url, files=multipart_form_data, params=params, timeout=self.timeout
        )

        # raise an error if the response is not as expected
        if response.status_code != 200:
            raise ValueError(
                f"HTTP request for prediction not successful. Status code: {response.status_code}"
            )

        body = response.json()

        if body["format_version"] != "0.2":
            logging.warning(
                "Using segmentation approach with unsupported version number!"
            )

        content = body["segmentation_data"][0]

        for detection in content:
            # label = detection['label']
            contour = np.array(detection["contour_coordinates"], dtype=np.float32)
            score = -1.0

            contours.append(Contour(contour, score, frame_id, -1))

        return contours

    @retry(requests.exceptions.ReadTimeout, tries=3)
    def predict_batch(self, frame_ids: List[int], images: List, params) -> Overlay:
        """Predict segmentation for a batch of frames

        Args:
            frame_ids (List[int]): ids of the frames
            images (List): numpy array for every image
            params ([type]): dictionary of additional parameters

        Raises:
            ValueError: [description]

        Returns:
            [Overlay]: An overlay containing the segmentation information for the images
        """
        contours = []

        binary_images = []

        for image in images:
            # convert image into a binary png stream
            byte_io = BytesIO()
            Image.fromarray(image).save(byte_io, "png")
            byte_io.seek(0)

            binary_images.append(byte_io)

        multipart_form_data = [
            ("files", ("data.png", bin_image, "image/png"))
            for bin_image in binary_images
        ]

        # send a request to the server
        response = requests.post(
            self.url, files=multipart_form_data, params=params, timeout=self.timeout
        )

        # raise an error if the response is not as expected
        if response.status_code != 200:
            raise ValueError(
                f"HTTP request for prediction not successful. Status code: {response.status_code}"
            )

        body = response.json()

        if body["format_version"] != "0.2":
            logging.warning(
                "Using segmentation approach with unsupported version number!"
            )

        for i, frame_id in enumerate(frame_ids):
            content = body["segmentation_data"][i]
            for detection in content:
                # label = detection['label']
                contour = np.array(detection["contour_coordinates"], dtype=np.float32)
                score = -1.0

                contours.append(Contour(contour, score, frame_id, -1))

        logging.debug("Finished batch prediction")

        return contours

    @staticmethod
    def parseContours(response_body) -> List[Contour]:
        pass
